/* Created by Wade Tregaskis on Fri Oct 04 2002. */

//  Copyright (c) 2006, Wade Tregaskis.  All rights reserved.
//  Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:
//    * Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
//    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
//    * Neither the name of Wade Tregaskis nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


#include <stdio.h>
#include <string.h>
#include <stdarg.h>


// -- DEFINES -- //
#define NAME_LENGTH_LIMIT 80
#define ADDRESS_LENGTH_LIMIT 80
#define CITY_LENGTH_LIMIT 40
#define RECORD_LIMIT 25
#define LINE_BUFFER_SIZE 80


// -- PSUEDO BOOLEAN SUPPORT -- //
#define true 1
#define false 0
typedef int bool;


// -- ENUMS -- //
enum stat {current, overdue, delinquent, credited};


// -- STRUCTS -- //
typedef struct date {
    unsigned int year, month, day;
} Date;

typedef struct rec {
    char name[NAME_LENGTH_LIMIT], address[ADDRESS_LENGTH_LIMIT], city[CITY_LENGTH_LIMIT];
    unsigned int accountNumber;
    enum stat status;
    float previousBalance, currentPayment, newBalance;
    Date paymentDate;
} Record;


// -- FUNCTION DECLARATIONS -- //
void listCustomers(Record list[], int count);
void listAllCustomers(Record list[], int count);
void listCertainCustomers(Record list[], int count, enum stat type);
bool addRecord(Record list[], int *count, int limit);
void displayRecord(Record list[], int index, int count);

char* getline(char* buffer, int maxLength);
void getDate(Date *item);
int getInteger(void);
unsigned int getUnsignedInteger(void);
double getDouble(void);
bool getBoolean(void);
inline void clearLine(FILE *io);

bool isBoolean(const char *data);
bool boolFromString(const char *data);
int caseInsensitiveStringCompare(const char *a, const char *b);

int showIndexedMenu(const char *title, int optionCount, ...);


// -- IMPLEMENTATION -- //

/* main(void)
     Simply displays the main menu until the user chooses to exit. */

int main(void) {
    Record customer[RECORD_LIMIT];
    int choice = 0, numberOfRecords = 0;

    do {
        choice = showIndexedMenu("Main Menu", 4, "List Customers", 1, "Add Customer", 2, NULL, 0, "Exit", 4);
        puts("");
        
        switch (choice) {
        case 1:
            listCustomers(customer, numberOfRecords);
            break;
        case 2:
            addRecord(customer, &numberOfRecords, RECORD_LIMIT);
            break;
        }
    } while (choice != 4);
    
    return 0;
}

/* listCustomers(Record list[], int count)
     Displays a submenu offering different listings of the customers in 'list'.  Continues to do so until the user chooses to return to the main menu.  'count' should be the number of records in 'list'. */

void listCustomers(Record list[], int count) {
    int choice = 0;

    if (count) {
        do {
            choice = showIndexedMenu("Customer Listing", 7, "List All", 1, "List Current", 2, "List Overdue", 3, "List Delinquent", 4, "List Credited", 5, NULL, 0, "Back", 6);
            puts("");

            switch (choice) {
            case 1:
                listAllCustomers(list, count);
                break;
            case 2:
                listCertainCustomers(list, count, current);
                break;
            case 3:
                listCertainCustomers(list, count, overdue);
                break;
            case 4:
                listCertainCustomers(list, count, delinquent);
                break;
            case 5:
                listCertainCustomers(list, count, credited);
                break;
            }
        } while (choice != 6);
    } else {
        puts("No customers on record.\n");
    }
}

/* listAllCustomers(Record list[], int count)
     Very simply displays every customer in 'list', where 'count' is the number of customers. */

void listAllCustomers(Record list[], int count) {
    int i;

    for (i = 0; i < count; ++i) {
        displayRecord(list, i, count);
    }

    puts("");
}

/* listCertainCustomers(Record list[], int count, enum stat type)
    Displays every customer in 'list', up to 'count', who's account status is 'type'. */

void listCertainCustomers(Record list[], int count, enum stat type) {
    int i, displayCount = 0;

    for (i = 0; i < count; ++i) {
        if (list[i].status == type) {
            ++displayCount;
            displayRecord(list, i, count);
        }
    }

    if (displayCount) {
        puts("");
    } else {
        puts("No customers of that type on record.\n");
    }
}

/* addRecord(Record list[], int *count, int limit)
     Appends a new record to the end of 'list', at index 'count', providing that is less than 'limit'.  Gathers all data for the new record from the user via stdin.  Returns true if the record was added successfully, false if not for any reason (e.g. limit reached).  Adjusts count appropriately to reflect the addition (or not) of a new record. */

bool addRecord(Record list[], int *count, int limit) {
    Record *holder;
    
    if (*count == limit) {
        puts("Unable to add any more customers - list is full!\n");
        return false;
    }

    holder = &list[*count];
    
    printf("New Customer Name:  ");
    getline(holder->name, NAME_LENGTH_LIMIT);
    printf("Address:  ");
    getline(holder->address, ADDRESS_LENGTH_LIMIT);
    printf("City/State:  ");
    getline(holder->city, CITY_LENGTH_LIMIT);
    printf("Account number:  ");
    holder->accountNumber = getUnsignedInteger();
    printf("Previous balance:  ");
    holder->previousBalance = getDouble();

    if (holder->previousBalance < 0) {
        printf("Are you sure the previous balance was $%5.2lf?  ", holder->previousBalance);

        if (!getBoolean()) {
            puts("");
            return false;
        }
    }
    
    printf("Current payment:  ");
    holder->currentPayment = getDouble();

    holder->newBalance = holder->previousBalance - holder->currentPayment;

    if (holder->currentPayment < 0) {
        printf("Are you sure the payment was $%5.2lf?  ", holder->currentPayment);

        if (!getBoolean()) {
            puts("");
            return false;
        }
    }
    
    if (holder->currentPayment > 0) {
        if (holder->newBalance < 0) {
            printf("That gives a new balance of $%5.2lf.  Are you sure this is correct?  ", holder->newBalance);

            if (!getBoolean()) {
                puts("");
                return false;
            }
        }

        printf("Payment date:  ");
        getDate(&(holder->paymentDate));
    }

    if (holder->newBalance < 0) {
        holder->status = credited;
    } else {
        if (holder->currentPayment > 0) {
            if (holder->currentPayment < (holder->previousBalance / 2)) {
                holder->status = overdue;
            } else {
                holder->status = current;
            }
        } else {
            holder->status = delinquent;
        }
    }
    
    ++(*count);

    puts("New customer successfully added.\n");

    return true;
}

/* displayRecord(Record list[], int index, int count)
     Displays the record in 'list' at 'index' to stdout, neatly formatted and appropriately arranged.  Does boundary checking with 'count', which should be the number of records in 'list'. */

void displayRecord(Record list[], int index, int count) {
    Record *holder;

    if (index < 0 || index >= count) {
        return;
    }

    holder = &(list[index]);

    printf("Name: %-44s Account number: %u\n", holder->name, holder->accountNumber);
    printf("Street: %s\nCity: %s\n\n", holder->address, holder->city);
    printf("Old balance: $%-10.2lf       Current payment: $%-5.2lf\n", holder->previousBalance, holder->currentPayment);
    printf("New balance: $%-10.2lf", holder->newBalance);

    if (holder->currentPayment > 0) {
        printf("       Payment date: %u/%u/%02u", (holder->paymentDate).day, (holder->paymentDate).month, (holder->paymentDate).year);
    }
    
    printf("\n\nAccount status: ");

    switch (holder->status) {
    case current:
        printf("Current");
        break;
    case overdue:
        printf("Overdue");
        break;
    case delinquent:
        printf("Delinquent");
        break;
    case credited:
        printf("Credited");
        break;
    }

    printf("\n\n");
}

/* getline(char* buffer, int maxLength)
     Gets an entire line from stdin and places it in buffer.  The behavior if the user enters more characters than 'maxLength' is undefined - excess data may be left in the buffer, or may be lost. */

char* getline(char* buffer, int maxLength) {
    char *temp;

#ifdef __USE_GNU
    getline(buffer, maxLength, stdin);
#else
    fgets(buffer, maxLength, stdin);

    temp = strchr(buffer, '\n');

    if (temp) {
        *temp = '\0';
    } else {
        buffer[maxLength - 1] = '\0';
    }
#endif
    
    return buffer;
}

/* getDate(Date *item)
     Has the user input a date of the format 'dd/mm/yy', and places the data into 'item'.  Performs some basic input validation, but should not be considered accurate. */

void getDate(Date *item) {
    bool happy = false;

    do {
        if ((scanf("%u/%u/%u", &(item->day), &(item->month), &(item->year)) == 3) && ((signed int)item->day > 0) && ((signed int)item->month > 0) && ((signed int)item->year >= 0)) {
            switch (item->month) {
            case 1:
            case 3:
            case 5:
            case 7:
            case 8:
            case 10:
            case 12:
                if (item->day < 32) {
                    happy = true;
                } else {
                    printf("There are only 31 days in that month, not %d.\n", item->day);
                }

                break;
            case 2:
                if (item->year % 4 == 0) {
                    if (item->day < 30) {
                        happy = true;
                    } else {
                        printf("There are only 29 days in February of a leap year, not %d.\n", item->day);
                    }
                } else {
                    if (item->day < 29) {
                        happy = true;
                    } else if (item->day == 29) {
                        printf("There are only 29 days in February in a leap year.  %02d is not a leap year.\n", item->year);
                    } else {
                        printf("There are only 28 days in February, not %d.\n", item->day);
                    }
                }

                break;
            case 4:
            case 6:
            case 9:
            case 11:
                if (item->day < 31) {
                    happy = true;
                } else {
                    printf("There are only 30 days in that month, not %d.\n", item->day);
                }

                break;
            default:
                puts("Invalid month - only values 1 to 12 inclusive are acceptable.");
            }
        }

        clearLine(stdin);
        
        if (!happy) {
            printf("Please enter a valid date (e.g. '2/3/02'):  ");
        }
    } while (!happy);
}

/* getInteger(void)
     Gets an integer value out of the user and returns it. */

int getInteger(void) {
    int temp;

    while (!scanf("%d", &temp)) {
        clearLine(stdin);
        printf("Please enter an integer value (e.g. '3', '1', '-23'):  ");
    }

    clearLine(stdin);
    
    return temp;
}

/* getUnsignedInteger(void)
     Gets a positive (unsigned) integer value out of the user and returns it. */

unsigned int getUnsignedInteger(void) {
    unsigned int temp;

    while (!scanf("%u", &temp) || ((signed int)temp < 0)) {
        clearLine(stdin);
        printf("Please enter a positive integer value (e.g. '3', '1', '23'):  ");
    }

    clearLine(stdin);

    return temp;
}

/* getDouble(void)
     Gets a floating point (double, specifically) value out of the user and returns it. */

double getDouble(void) {
    double temp;

    while (!scanf("%lf", &temp)) {
        clearLine(stdin);
        printf("Please enter a decimal value (e.g. '2', '3.5', '-283.58'):  ");
    }

    clearLine(stdin);
    
    return temp;
}

/* getBoolean(void)
     Gets a boolean value out of the user and returns it.  Accepts a variety of formats, from simple 'y' to 'yes' to '1' or 't', etc etc. */

bool getBoolean(void) {
    char temp[LINE_BUFFER_SIZE];

    while (1) {
        getline(temp, LINE_BUFFER_SIZE);

        if (isBoolean(temp)) {
            return boolFromString(temp);
        } else {
            printf("Please enter a yes or no answer (e.g. 'yes', 'y', 'no'):  ");
        }
    }
}

/* isBoolean(const char *data)
     Returns true if the string in 'data' can be taken as a true or false boolean false.  Returns false otherwise.  Allows for a variety of inputs, from single character to full word. */

bool isBoolean(const char *data) {
    char smallTemp[2] = {'\0', '\0'};

    if (data) {
        if (strlen(data) == 1) {
            smallTemp[0] = toupper(data[0]);

            return (strpbrk(smallTemp, "YNTF10+-") != NULL);
        } else {
            return (caseInsensitiveStringCompare(data, "yes") == 0 ||
                    caseInsensitiveStringCompare(data, "no") == 0 ||
                    caseInsensitiveStringCompare(data, "true") == 0 ||
                    caseInsensitiveStringCompare(data, "false") == 0);
        }
    } else {
        return false;
    }
}

/* boolFromString(const char *data)
     Determines what boolean value, true or false, the string in 'data' represents, and returns that value.  In the event that no value can be identified from 'data', returns false. */

bool boolFromString(const char *data) {
    if (strlen(data)) {
        switch (toupper(data[0])) {
            case 'Y': // Yes
                return true;
            case 'N': // No/Negative
                return false;
            case '1': // 1
                return true;
            case '0': // 0
                return false;
            case 'T': // True
                return true;
            case 'F': // False
                return false;
            case '+': // +
                return true;
            case '-': // -
                return false;
        }
    }

    return false;
}

/* caseInsensitiveStringCompare(const char *a, const char *b)
     Same functionality as strcmp, except it ignores case, as the name suggests.  Thus 'c' is the same as 'C', for example. */

int caseInsensitiveStringCompare(const char *a, const char *b) {
    int offset, end, lengthA, lengthB, charA, charB;

    if (a == b) {
        return 0;
    }

    lengthA = strlen(a);
    lengthB = strlen(b);

    end = (lengthA < lengthB) ? lengthA : lengthB;

    for(offset = 0; offset < end; ++offset) {
        charA = toupper(a[offset]);
        charB = toupper(b[offset]);

        if (charA != charB) {
            return (charA < charB) ? -1 : 1;
        }
    }

    if (lengthA == lengthB) {
        return 0;
    } else {
        return (lengthA < lengthB) ? -1 : 1;
    }
}

/* showIndexedMenu(const char *title, int optionCount, ...)
     Draws a generic menu with the title of 'title', and containing 'optionCount'
     number of items.  After this, each option is passed as a pair of arguments -
     first a const char* to the option name.  Then an integer which is returned if
     that option is choosen.
       To display a blank link in the menu as a whitespace separator, pass NULL
     as the item name and 0 as the return value.  If you pass NULL but a return
     value of anything else, that entry will be ignored. */

int showIndexedMenu(const char *title, int optionCount, ...) {
    va_list arg;
    int *menuList, realCount = 0;
    unsigned int choice = 0, i;
    const char *name = NULL;

    if (optionCount > 0) {
        menuList = (int*)malloc(sizeof(int) * optionCount);

        printf("--<== %s ==>--\n\n", title);

        va_start(arg, optionCount);

        for(i = 0; i < optionCount; ++i) {
            name = va_arg(arg, const char*);
            menuList[realCount] = va_arg(arg, int);

            if (name) {
                printf("  %d. %s\n", realCount + 1, name);
                ++realCount;
            } else {
                if (menuList[realCount] == 0) {
                    puts("");
                }
            }
        }

        va_end(arg);

        puts("");

        while (1) {
            printf("Selection: ");
            choice = getUnsignedInteger();

            if ((choice < 1) || (choice > realCount)) {
                printf("Invalid selection - please enter a number between 1 and %d.\n", realCount);
            } else {
                break;
            }
        }

        realCount = menuList[choice - 1];
        free(menuList);

        return realCount;
    } else {
        return 0;
    }
}

/* clearLine(FILE *io)
     Reads all characters up to and including the next newline in 'io'. */

inline void clearLine(FILE *io) {
    while (fgetc(io) != '\n') {}
}
